/*
Copyright (c) 2003 Forward Computing and Control Pty. Ltd.
NSW. Australia,  www.forward.com.au
All rights reserved.

Redistribution of the source of this package, with or without
modification, is NOT permitted.

Redistribution of the compiled java .class files of this package, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions in compiled java .class form must reproduce the above
   copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the distribution.

2. If the software is modified, the compiled java .class distribution must
   include, in the documentation and/or other materials provided with the 
   distribution, a statement that the software has been modified and
   identify the person who was responsable for the modifications.  

3. Neither the name of the author nor Forward Computing and Control may be 
   used to endorse or promote products derived from this software
   without specific prior written permission.

4.  All advertising materials mentioning features or use of this
    software must display the following acknowledgment: 
    "This product includes software developed by Matthew Ford
     and Forward Computing and Control Pty. Ltd. 
    (http://www.forward.com.au)."


THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
*/
package au.com.forward.logging;

import java.util.*;
import java.util.logging.*;
import java.io.File;

/**
 *  This class provides logging initialization and debugging.<br>
 *  Typical use for logging is <pre><code>
 *  import au.com.forward.logging.*;
 *  static {
 *        // set up logging  also redirects System.err and System.out
 *        // throws a LoggingException if problem with setting up Java logging
 *        // put all log files in user's home dir
 *        // roll log to .old when it exceed 100K
 *        String userHome = System.getProperty("user.home",".")
 *                         +System.getProperty("file.separator","/");
 *        Logging.initializeLogging(userHome+"StatusApplication.log",
 *                "Log File for StatusApplication " + new Date(),true, 100);
 *        // redirect System.out as well as System.err and append to existing log
 *   }
 *</pre></code> <p>
 *
 *  If you want your application to continue even if there is an error setting
 *  up Java Logging then use the following code<br>
 *  but usually you would want the application to stop if logging fails.<br>
 *  Note: The System.err and System.out redirection are called before
 *  initializing Java Logging so they are still in effect even if logging fails.
 *  <code><pre>
 *  static {
 *       // set up logging  also redirects System.err and System.out
 *       // throws a LoggingException if problem with setting up Java logging
 *       // put all log files in user's home dir
 *        // roll log to .old when it exceed 100K
 *         String userHome = System.getProperty("user.home",".")
 *             +System.getProperty("file.separator","/");
 *       try {
 *           Logging.initializeLogging(userHome+"StatusApplication.log",
 *              "Log File for StatusApplication " + new Date(),true, 100);
 *           // redirect System.out as well as System.err and append to existing log
 *       } catch (LoggingException lex) {
 *           System.err.println(RobustFormatter.toString(lex));
 *       }
 *   }
 *</pre></code> <p>
 *
 * If you are using Log4J or some other non-Java logging package,
 *  then you should use 
 * {@link LogStdStreams#initializeErrorLogging(String, String ,boolean, int)}
 * instead to capture the console output from the logger (and any erronous System.out.println's)
 * <p> 

 <h3>Debugging Logging Settings</h3>
 * For debugging your logging settings you can use 
 * <pre><code>   System.out.println({@link #currentConfiguration() Logging.currentConfiguration()});
 *</pre></code>which will dump the current logging setting in your log file.<p>
 <h3>Logging Properties File</h3>
 
   A sample properties file is <pre><code>
 ############################################################
 #   Logging Configuration File
 #
 # To use this file specify its filename
 # with the java.util.logging.config.file system property.
 #
 # For example:
 #     java -Djava.util.logging.config.file=logging.properties  au.com.forward.ExampleLoggingApplication
 #
 ############################################################
 ############################################################
 #   Global properties
 ############################################################
 # "handlers" specifies a comma separated list of log Handler
 # classes.  These handlers will be installed during VM startup.
 # Note that these classes must be on the system classpath.
 # Add a FileHandler
 handlers=java.util.logging.FileHandler
 # Default global logging level.
 # This specifies which kinds of events are logged across
 # all loggers.  For any given facility this global level
 # can be overridden by a facility specific level
 .level = SEVERE
 ############################################################
 # Handler specific properties.
 # Describes specific configuration info for Handlers.
 ############################################################
 # default file output is in user's home directory.
 java.util.logging.FileHandler.pattern = %h/applicationLog%u.xml
 java.util.logging.FileHandler.formatter = au.com.forward.logging.LoggingXMLFormatter
 ############################################################
 # Facility specific properties.
 # Provides extra control for each logger.
 ############################################################
 # messages:
 au.com.forward.ExampleLoggingApplication.level = FINEST
 </code></pre>
 *
 *@author     Matthew Ford
 *@created    October 9, 2003
 */
public class Logging {

  /**
   * The version number  1.3 
	 */
	public static final String VERSION = "1.3";
	
	/** The version date  23nd Nov. 2004
	 *
	 */
	public static final String VERSION_DATE = "23nd Nov. 2004";
	
	/**
	 * private constructor as all methods are static
	 */
	private Logging() {
	}

  /**
   *  Initializes the Log file and sets up the default Console Handler
   *
   *@param  logFileName           the name of the log file
   *@param  initialStr            the heading string to write to the log file when opened
   *@param  logStdOut             true if System.Out to redirected to log file also
   *@param  append                true if to append to existing log, false to rewrite new
   *  log.
   *@exception  LoggingException  thrown initializeErrorLogging() has already
   *  been called.
   */
  public static void initializeLogging(String logFileName, String initialStr,
      boolean logStdOut, boolean append) throws LoggingException {
    // set up logging for errors
    LogStdStreams.initializeErrorLogging(logFileName, initialStr, logStdOut, append);
    initializeLogging();
  }

  /**
   *  Initializes a rolling Log file and sets up the default Console Handler
   *
   *@param  logFileName           the name of the log file
   *@param  initialStr            the heading string to write to the log file when opened
   *@param  logStdOut             true if System.Out to redirected to log file also
   *@param  maxSizeInKbytes  roll current file to .old if it exceeds this size on
   * startup.  Any existing .old log file is deleted. 
   *@exception  LoggingException  thrown initializeErrorLogging() has already
   *  been called.
   */
  public static void initializeLogging(String logFileName, String initialStr,
      boolean logStdOut, int maxSizeInKbytes) throws LoggingException {
    // set up logging for errors
    LogStdStreams.initializeErrorLogging(logFileName, initialStr, logStdOut, maxSizeInKbytes);
    initializeLogging();
  }


  /**
   *  Sets up the default ConsoleHandler without opening a log file
   *
   *@exception  LoggingException  thrown initializeErrorLogging() has already
   *  been called.
   */
  public static void initializeLogging() throws LoggingException {
    try {
      Handler[] handlers = Logger.getLogger("").getHandlers();
      boolean foundConsoleHandler = false;
      for (int index = 0; index < handlers.length; index++) {
        // set console handler to SEVERE
        if (handlers[index] instanceof ConsoleHandler) {
          handlers[index].setLevel(Level.SEVERE);
          handlers[index].setFormatter(new LoggingSimpleFormatter());
          foundConsoleHandler = true;
        }
      }
      if (!foundConsoleHandler) {
        // no console handler found
        System.err.println("No consoleHandler found, adding one.");
        ConsoleHandler consoleHandler = new ConsoleHandler();
        consoleHandler.setLevel(Level.SEVERE);
        consoleHandler.setFormatter(new LoggingSimpleFormatter());
        Logger.getLogger("").addHandler(consoleHandler);
      }
      // test the loggingXMLFormatter is on the path
      new LoggingXMLFormatter();
    } catch (Throwable t) {
      throw new LoggingException("Error initializing logging\n" + t.getMessage(), t);
    }
  }


  /**
   *  Returns the current logging configuration for debugging. <p>
	 *  Loggers created in classes are not visable in the current configuration until their class
	 *  is loaded.  This means the best place to call<br>
   * &nbsp; &nbsp; System.out.println(Logging.currentConfiguration());<br>
	 * is at the end of the application main() or in the exit code after all the classes have
   * been loaded.
   *
   *@return    a string describing the current logging configuration
   */
  public static String currentConfiguration() {
    try {
      StringBuffer rtn = new StringBuffer(1024);
      rtn.append("-- Logging configuration --\n");
			
      String classPath = System.getProperty("java.class.path");
      rtn.append(" The classpath is '"+classPath+"'\n");
			
      String loggingConfigClass = System.getProperty("java.util.logging.config.class");
      String loggingConfigFile = System.getProperty("java.util.logging.config.file");
			boolean configClassOK = false;
      if (loggingConfigClass == null) {
        rtn.append(" No java.util.logging.config.class class is set.\n");
      } else {
        rtn.append("java.util.logging.config.class is set to '" + loggingConfigClass + "'\n");
				try {
					Class c = Class.forName(loggingConfigClass);
					c.newInstance();
					rtn.append("   This class was loaded and a new instance was sucessfully created\n");
					configClassOK = true;
				} catch (Throwable t) {
					rtn.append("   An error occured creating this class.  The error was:\n");
					rtn.append(RobustFormatter.toString(t)+"\n");
				}						
      }
			
      if (loggingConfigFile == null) {
        rtn.append(" No java.util.logging.config.file file is set.\n");
      } else {
        rtn.append("java.util.logging.config.file is set to '" + loggingConfigFile + "'\n");
        File loggingFile = new File(loggingConfigFile);
        rtn.append("      '"+loggingFile.getAbsolutePath() + "'\n");
        if (loggingFile.exists() && loggingFile.isFile()) {
          rtn.append("   and this file exists.\n");
        } else {
          rtn.append("   and this file does NOT EXIST.\n");
        }
				if (loggingConfigClass != null) {
					if (configClassOK) {
						rtn.append("  This file is ignored because java.util.logging.config.class is set (see above)"+ "\n");
						rtn.append("    and it takes precedence.\n");
					} else {
						rtn.append("  The java.util.logging.config.class failed to load (see above)"+ "\n");
						rtn.append("    so the settings in the java.util.logging.config.file take effect.\n");
					}	
				}	
      }
      Handler[] handlers = Logger.getLogger("").getHandlers();
      rtn.append("\n    There are " + handlers.length + " logging handlers :-\n");
      listHandlers(handlers,rtn,"");
      rtn.append("\n    Currently known loggers are:-\n");
      LogManager logManager = LogManager.getLogManager(); 
      Enumeration en = logManager.getLoggerNames();
      for (; en.hasMoreElements(); ) {
        String loggerName = (String)en.nextElement();
        Logger logger = logManager.getLogger(loggerName);
        Filter filter = logger.getFilter();
        
        handlers = logger.getHandlers();
        Logger handlerParent = null;
        if (handlers.length == 0) {
          handlerParent = logger;
          while ((handlerParent != null) && 
               ((handlers = handlerParent.getHandlers()).length==0) ) {
            handlerParent = handlerParent.getParent();
          }
        } 
        
        Level level = logger.getLevel();
        Logger levelParent = null;
        if (level == null) {
          levelParent = logger;
          while ((levelParent != null) && 
               ((level=levelParent.getLevel())== null) ) {
            levelParent = levelParent.getParent();
          }
        }
        
        String resourceBundleName = logger.getResourceBundleName();
        Logger resourceParent = null;
        if (resourceBundleName == null) {
          resourceParent = logger;
          while ((resourceParent != null) && 
               ((resourceBundleName=resourceParent.getResourceBundleName())== null) ) {
            resourceParent = resourceParent.getParent();
          }
        } 
        
        rtn.append("LoggerName:'"+loggerName+"'\n");
        rtn.append("    Level:"+level+"  "
                   +((levelParent==null)?"":"set by parent:'"+levelParent.getName()+"'")+"\n");
        rtn.append("    Filter:"+((filter==null)?"none set":"'"+filter.getClass().getName()+"'")+"\n");
        rtn.append("    ResourceBundleName:"
                   +((resourceBundleName==null)?" none set ":"'"+resourceBundleName+"'  ")
                   +((resourceParent==null)?"":"set by parent:'"+resourceParent.getName()+"'")+"\n");
        rtn.append("    Handlers: "
                   +((handlerParent==null)?"":"set by parent:'"+handlerParent.getName()+"'")+"\n");
        listHandlers(handlers,rtn,"             ");
        rtn.append("\n");
      }
      rtn.append("NOTE: More loggers may be added as classes are loaded by the JVM.\n");
      return rtn.toString();
    } catch (Throwable t) {
      return "Error getting current logging configuration\n"+RobustFormatter.toString(t);
    } 
  }

  
  private static StringBuffer listHandlers(Handler[] handlers, StringBuffer buffer, String indent) {
    if (indent == null) {
      indent = "";
    }
    for (int index = 0; index < handlers.length; index++) {
      buffer.append(indent+"Handler:" + handlers[index].getClass().getName() +
          " Level:" + handlers[index].getLevel().toString() +
          " Formatter:" + handlers[index].getFormatter().getClass().getName() + "\n");
    }
    return buffer;
  }

}

